@@ -46,6 +46,10 @@ import (
 	"github.com/prometheus/prometheus/util/testutil"
 )
 
+func TestMain(m *testing.M) {
+	testutil.TolerantVerifyLeak(m)
+}
+
 func TestNewScrapePool(t *testing.T) {
 	var (
 		app   = &nopAppendable{}
@@ -134,17 +138,36 @@ func TestDiscoveredLabelsUpdate(t *testing.T) {
 }
 
 type testLoop struct {
-	startFunc func(interval, timeout time.Duration, errc chan<- error)
-	stopFunc  func()
+	startFunc    func(interval, timeout time.Duration, errc chan<- error)
+	stopFunc     func()
+	forcedErr    error
+	forcedErrMtx sync.Mutex
+	runOnce      bool
 }
 
 func (l *testLoop) run(interval, timeout time.Duration, errc chan<- error) {
+	if l.runOnce {
+		panic("loop must be started only once")
+	}
+	l.runOnce = true
 	l.startFunc(interval, timeout, errc)
 }
 
 func (l *testLoop) disableEndOfRunStalenessMarkers() {
 }
 
+func (l *testLoop) setForcedError(err error) {
+	l.forcedErrMtx.Lock()
+	defer l.forcedErrMtx.Unlock()
+	l.forcedErr = err
+}
+
+func (l *testLoop) getForcedError() error {
+	l.forcedErrMtx.Lock()
+	defer l.forcedErrMtx.Unlock()
+	return l.forcedErr
+}
+
 func (l *testLoop) stop() {
 	l.stopFunc()
 }
@@ -297,6 +320,128 @@ func TestScrapePoolReload(t *testing.T) {
 	testutil.Equals(t, numTargets, len(sp.loops), "Unexpected number of stopped loops after reload")
 }
 
+func TestScrapePoolTargetLimit(t *testing.T) {
+	var wg sync.WaitGroup
+	// On starting to run, new loops created on reload check whether their preceding
+	// equivalents have been stopped.
+	newLoop := func(opts scrapeLoopOptions) loop {
+		wg.Add(1)
+		l := &testLoop{
+			startFunc: func(interval, timeout time.Duration, errc chan<- error) {
+				wg.Done()
+			},
+			stopFunc: func() {},
+		}
+		return l
+	}
+	sp := &scrapePool{
+		appendable:    &nopAppendable{},
+		activeTargets: map[uint64]*Target{},
+		loops:         map[uint64]loop{},
+		newLoop:       newLoop,
+		logger:        nil,
+		client:        http.DefaultClient,
+	}
+
+	var tgs = []*targetgroup.Group{}
+	for i := 0; i < 50; i++ {
+		tgs = append(tgs,
+			&targetgroup.Group{
+				Targets: []model.LabelSet{
+					{model.AddressLabel: model.LabelValue(fmt.Sprintf("127.0.0.1:%d", 9090+i))},
+				},
+			},
+		)
+	}
+
+	var limit uint
+	reloadWithLimit := func(l uint) {
+		limit = l
+		testutil.Ok(t, sp.reload(&config.ScrapeConfig{
+			ScrapeInterval: model.Duration(3 * time.Second),
+			ScrapeTimeout:  model.Duration(2 * time.Second),
+			TargetLimit:    l,
+		}))
+	}
+
+	var targets int
+	loadTargets := func(n int) {
+		targets = n
+		sp.Sync(tgs[:n])
+	}
+
+	validateIsRunning := func() {
+		wg.Wait()
+		for _, l := range sp.loops {
+			testutil.Assert(t, l.(*testLoop).runOnce, "loop should be running")
+		}
+	}
+
+	validateErrorMessage := func(shouldErr bool) {
+		for _, l := range sp.loops {
+			lerr := l.(*testLoop).getForcedError()
+			if shouldErr {
+				testutil.Assert(t, lerr != nil, "error was expected for %d targets with a limit of %d", targets, limit)
+				testutil.Equals(t, fmt.Sprintf("target_limit exceeded (number of targets: %d, limit: %d)", targets, limit), lerr.Error())
+			} else {
+				testutil.Equals(t, nil, lerr)
+			}
+		}
+	}
+
+	reloadWithLimit(0)
+	loadTargets(50)
+	validateIsRunning()
+
+	// Simulate an initial config with a limit.
+	sp.config.TargetLimit = 30
+	limit = 30
+	loadTargets(50)
+	validateIsRunning()
+	validateErrorMessage(true)
+
+	reloadWithLimit(50)
+	validateIsRunning()
+	validateErrorMessage(false)
+
+	reloadWithLimit(40)
+	validateIsRunning()
+	validateErrorMessage(true)
+
+	loadTargets(30)
+	validateIsRunning()
+	validateErrorMessage(false)
+
+	loadTargets(40)
+	validateIsRunning()
+	validateErrorMessage(false)
+
+	loadTargets(41)
+	validateIsRunning()
+	validateErrorMessage(true)
+
+	reloadWithLimit(51)
+	validateIsRunning()
+	validateErrorMessage(false)
+
+	tgs = append(tgs,
+		&targetgroup.Group{
+			Targets: []model.LabelSet{
+				{model.AddressLabel: model.LabelValue("127.0.0.1:1090")},
+			},
+		},
+		&targetgroup.Group{
+			Targets: []model.LabelSet{
+				{model.AddressLabel: model.LabelValue("127.0.0.1:1090")},
+			},
+		},
+	)
+
+	sp.Sync(tgs)
+	validateIsRunning()
+	validateErrorMessage(false)
+}
+
 func TestScrapePoolAppender(t *testing.T) {
 	cfg := &config.ScrapeConfig{}
 	app := &nopAppendable{}
@@ -308,7 +453,7 @@ func TestScrapePoolAppender(t *testing.T) {
 	appl, ok := loop.(*scrapeLoop)
 	testutil.Assert(t, ok, "Expected scrapeLoop but got %T", loop)
 
-	wrapped := appl.appender()
+	wrapped := appl.appender(context.Background())
 
 	tl, ok := wrapped.(*timeLimitAppender)
 	testutil.Assert(t, ok, "Expected timeLimitAppender but got %T", wrapped)
@@ -323,7 +468,7 @@ func TestScrapePoolAppender(t *testing.T) {
 	appl, ok = loop.(*scrapeLoop)
 	testutil.Assert(t, ok, "Expected scrapeLoop but got %T", loop)
 
-	wrapped = appl.appender()
+	wrapped = appl.appender(context.Background())
 
 	sl, ok := wrapped.(*limitAppender)
 	testutil.Assert(t, ok, "Expected limitAppender but got %T", wrapped)
@@ -372,6 +517,54 @@ func TestScrapePoolRaces(t *testing.T) {
 	sp.stop()
 }
 
+func TestScrapePoolScrapeLoopsStarted(t *testing.T) {
+	var wg sync.WaitGroup
+	newLoop := func(opts scrapeLoopOptions) loop {
+		wg.Add(1)
+		l := &testLoop{
+			startFunc: func(interval, timeout time.Duration, errc chan<- error) {
+				wg.Done()
+			},
+			stopFunc: func() {},
+		}
+		return l
+	}
+	sp := &scrapePool{
+		appendable:    &nopAppendable{},
+		activeTargets: map[uint64]*Target{},
+		loops:         map[uint64]loop{},
+		newLoop:       newLoop,
+		logger:        nil,
+		client:        http.DefaultClient,
+	}
+
+	tgs := []*targetgroup.Group{
+		{
+			Targets: []model.LabelSet{
+				{model.AddressLabel: model.LabelValue("127.0.0.1:9090")},
+			},
+		},
+		{
+			Targets: []model.LabelSet{
+				{model.AddressLabel: model.LabelValue("127.0.0.1:9090")},
+			},
+		},
+	}
+
+	testutil.Ok(t, sp.reload(&config.ScrapeConfig{
+		ScrapeInterval: model.Duration(3 * time.Second),
+		ScrapeTimeout:  model.Duration(2 * time.Second),
+	}))
+	sp.Sync(tgs)
+
+	testutil.Equals(t, 1, len(sp.loops))
+
+	wg.Wait()
+	for _, l := range sp.loops {
+		testutil.Assert(t, l.(*testLoop).runOnce, "loop should be running")
+	}
+}
+
 func TestScrapeLoopStopBeforeRun(t *testing.T) {
 	scraper := &testScraper{}
 
@@ -434,7 +627,7 @@ func TestScrapeLoopStop(t *testing.T) {
 		signal   = make(chan struct{}, 1)
 		appender = &collectResultAppender{}
 		scraper  = &testScraper{}
-		app      = func() storage.Appender { return appender }
+		app      = func(ctx context.Context) storage.Appender { return appender }
 	)
 
 	sl := newScrapeLoop(context.Background(),
@@ -455,9 +648,10 @@ func TestScrapeLoopStop(t *testing.T) {
 		numScrapes++
 		if numScrapes == 2 {
 			go sl.stop()
+			<-sl.ctx.Done()
 		}
 		w.Write([]byte("metric_a 42\n"))
-		return nil
+		return ctx.Err()
 	}
 
 	go func() {
@@ -499,7 +693,7 @@ func TestScrapeLoopRun(t *testing.T) {
 		errc   = make(chan error)
 
 		scraper = &testScraper{}
-		app     = func() storage.Appender { return &nopAppender{} }
+		app     = func(ctx context.Context) storage.Appender { return &nopAppender{} }
 	)
 
 	ctx, cancel := context.WithCancel(context.Background())
@@ -591,6 +785,57 @@ func TestScrapeLoopRun(t *testing.T) {
 	}
 }
 
+func TestScrapeLoopForcedErr(t *testing.T) {
+	var (
+		signal = make(chan struct{}, 1)
+		errc   = make(chan error)
+
+		scraper = &testScraper{}
+		app     = func(ctx context.Context) storage.Appender { return &nopAppender{} }
+	)
+
+	ctx, cancel := context.WithCancel(context.Background())
+	sl := newScrapeLoop(ctx,
+		scraper,
+		nil, nil,
+		nopMutator,
+		nopMutator,
+		app,
+		nil,
+		0,
+		true,
+	)
+
+	forcedErr := fmt.Errorf("forced err")
+	sl.setForcedError(forcedErr)
+
+	scraper.scrapeFunc = func(context.Context, io.Writer) error {
+		t.Fatalf("should not be scraped")
+		return nil
+	}
+
+	go func() {
+		sl.run(time.Second, time.Hour, errc)
+		signal <- struct{}{}
+	}()
+
+	select {
+	case err := <-errc:
+		if err != forcedErr {
+			t.Fatalf("Expected forced error but got: %s", err)
+		}
+	case <-time.After(3 * time.Second):
+		t.Fatalf("Expected forced error but got none")
+	}
+	cancel()
+
+	select {
+	case <-signal:
+	case <-time.After(5 * time.Second):
+		t.Fatalf("Scrape not stopped")
+	}
+}
+
 func TestScrapeLoopMetadata(t *testing.T) {
 	var (
 		signal  = make(chan struct{})
@@ -605,14 +850,14 @@ func TestScrapeLoopMetadata(t *testing.T) {
 		nil, nil,
 		nopMutator,
 		nopMutator,
-		func() storage.Appender { return nopAppender{} },
+		func(ctx context.Context) storage.Appender { return nopAppender{} },
 		cache,
 		0,
 		true,
 	)
 	defer cancel()
 
-	slApp := sl.appender()
+	slApp := sl.appender(ctx)
 	total, _, _, err := sl.append(slApp, []byte(`# TYPE test_metric counter
 # HELP test_metric some help text
 # UNIT test_metric metric
@@ -648,22 +893,20 @@ func TestScrapeLoopSeriesAdded(t *testing.T) {
 	s := teststorage.New(t)
 	defer s.Close()
 
-	app := s.Appender()
-
 	ctx, cancel := context.WithCancel(context.Background())
 	sl := newScrapeLoop(ctx,
 		&testScraper{},
 		nil, nil,
 		nopMutator,
 		nopMutator,
-		func() storage.Appender { return app },
+		s.Appender,
 		nil,
 		0,
 		true,
 	)
 	defer cancel()
 
-	slApp := sl.appender()
+	slApp := sl.appender(ctx)
 	total, added, seriesAdded, err := sl.append(slApp, []byte("test_metric 1\n"), "", time.Time{})
 	testutil.Ok(t, err)
 	testutil.Ok(t, slApp.Commit())
@@ -671,7 +914,7 @@ func TestScrapeLoopSeriesAdded(t *testing.T) {
 	testutil.Equals(t, 1, added)
 	testutil.Equals(t, 1, seriesAdded)
 
-	slApp = sl.appender()
+	slApp = sl.appender(ctx)
 	total, added, seriesAdded, err = sl.append(slApp, []byte("test_metric 1\n"), "", time.Time{})
 	testutil.Ok(t, slApp.Commit())
 	testutil.Ok(t, err)
@@ -685,7 +928,7 @@ func TestScrapeLoopRunCreatesStaleMarkersOnFailedScrape(t *testing.T) {
 	var (
 		signal  = make(chan struct{}, 1)
 		scraper = &testScraper{}
-		app     = func() storage.Appender { return appender }
+		app     = func(ctx context.Context) storage.Appender { return appender }
 	)
 
 	ctx, cancel := context.WithCancel(context.Background())
@@ -738,7 +981,7 @@ func TestScrapeLoopRunCreatesStaleMarkersOnParseFailure(t *testing.T) {
 	var (
 		signal     = make(chan struct{}, 1)
 		scraper    = &testScraper{}
-		app        = func() storage.Appender { return appender }
+		app        = func(ctx context.Context) storage.Appender { return appender }
 		numScrapes = 0
 	)
 
@@ -793,13 +1036,11 @@ func TestScrapeLoopCache(t *testing.T) {
 	s := teststorage.New(t)
 	defer s.Close()
 
-	sapp := s.Appender()
-
-	appender := &collectResultAppender{next: sapp}
+	appender := &collectResultAppender{}
 	var (
 		signal  = make(chan struct{}, 1)
 		scraper = &testScraper{}
-		app     = func() storage.Appender { return appender }
+		app     = func(ctx context.Context) storage.Appender { appender.next = s.Appender(ctx); return appender }
 	)
 
 	ctx, cancel := context.WithCancel(context.Background())
@@ -867,13 +1108,13 @@ func TestScrapeLoopCacheMemoryExhaustionProtection(t *testing.T) {
 	s := teststorage.New(t)
 	defer s.Close()
 
-	sapp := s.Appender()
+	sapp := s.Appender(context.Background())
 
 	appender := &collectResultAppender{next: sapp}
 	var (
 		signal  = make(chan struct{}, 1)
 		scraper = &testScraper{}
-		app     = func() storage.Appender { return appender }
+		app     = func(ctx context.Context) storage.Appender { return appender }
 	)
 
 	ctx, cancel := context.WithCancel(context.Background())
@@ -988,7 +1229,7 @@ func TestScrapeLoopAppend(t *testing.T) {
 			func(l labels.Labels) labels.Labels {
 				return mutateReportSampleLabels(l, discoveryLabels)
 			},
-			func() storage.Appender { return app },
+			func(ctx context.Context) storage.Appender { return app },
 			nil,
 			0,
 			true,
@@ -996,7 +1237,7 @@ func TestScrapeLoopAppend(t *testing.T) {
 
 		now := time.Now()
 
-		slApp := sl.appender()
+		slApp := sl.appender(context.Background())
 		_, _, _, err := sl.append(slApp, []byte(test.scrapeLabels), "", now)
 		testutil.Ok(t, err)
 		testutil.Ok(t, slApp.Commit())
@@ -1029,7 +1270,7 @@ func TestScrapeLoopAppendCacheEntryButErrNotFound(t *testing.T) {
 		nil, nil, nil,
 		nopMutator,
 		nopMutator,
-		func() storage.Appender { return app },
+		func(ctx context.Context) storage.Appender { return app },
 		nil,
 		0,
 		true,
@@ -1049,7 +1290,7 @@ func TestScrapeLoopAppendCacheEntryButErrNotFound(t *testing.T) {
 	sl.cache.addRef(mets, fakeRef, lset, hash)
 	now := time.Now()
 
-	slApp := sl.appender()
+	slApp := sl.appender(context.Background())
 	_, _, _, err := sl.append(slApp, []byte(metric), "", now)
 	testutil.Ok(t, err)
 	testutil.Ok(t, slApp.Commit())
@@ -1078,7 +1319,7 @@ func TestScrapeLoopAppendSampleLimit(t *testing.T) {
 			return l
 		},
 		nopMutator,
-		func() storage.Appender { return app },
+		func(ctx context.Context) storage.Appender { return app },
 		nil,
 		0,
 		true,
@@ -1092,7 +1333,7 @@ func TestScrapeLoopAppendSampleLimit(t *testing.T) {
 	beforeMetricValue := beforeMetric.GetCounter().GetValue()
 
 	now := time.Now()
-	slApp := sl.appender()
+	slApp := sl.appender(context.Background())
 	total, added, seriesAdded, err := sl.append(app, []byte("metric_a 1\nmetric_b 1\nmetric_c 1\n"), "", now)
 	if err != errSampleLimit {
 		t.Fatalf("Did not see expected sample limit error: %s", err)
@@ -1123,7 +1364,7 @@ func TestScrapeLoopAppendSampleLimit(t *testing.T) {
 	testutil.Equals(t, want, resApp.rolledbackResult, "Appended samples not as expected")
 
 	now = time.Now()
-	slApp = sl.appender()
+	slApp = sl.appender(context.Background())
 	total, added, seriesAdded, err = sl.append(slApp, []byte("metric_a 1\nmetric_b 1\nmetric_c{deleteme=\"yes\"} 1\nmetric_d 1\nmetric_e 1\nmetric_f 1\nmetric_g 1\nmetric_h{deleteme=\"yes\"} 1\nmetric_i{deleteme=\"yes\"} 1\n"), "", now)
 	if err != errSampleLimit {
 		t.Fatalf("Did not see expected sample limit error: %s", err)
@@ -1141,27 +1382,25 @@ func TestScrapeLoop_ChangingMetricString(t *testing.T) {
 	s := teststorage.New(t)
 	defer s.Close()
 
-	app := s.Appender()
-
-	capp := &collectResultAppender{next: app}
+	capp := &collectResultAppender{}
 
 	sl := newScrapeLoop(context.Background(),
 		nil, nil, nil,
 		nopMutator,
 		nopMutator,
-		func() storage.Appender { return capp },
+		func(ctx context.Context) storage.Appender { capp.next = s.Appender(ctx); return capp },
 		nil,
 		0,
 		true,
 	)
 
 	now := time.Now()
-	slApp := sl.appender()
+	slApp := sl.appender(context.Background())
 	_, _, _, err := sl.append(slApp, []byte(`metric_a{a="1",b="1"} 1`), "", now)
 	testutil.Ok(t, err)
 	testutil.Ok(t, slApp.Commit())
 
-	slApp = sl.appender()
+	slApp = sl.appender(context.Background())
 	_, _, _, err = sl.append(slApp, []byte(`metric_a{b="1",a="1"} 2`), "", now.Add(time.Minute))
 	testutil.Ok(t, err)
 	testutil.Ok(t, slApp.Commit())
@@ -1189,19 +1428,19 @@ func TestScrapeLoopAppendStaleness(t *testing.T) {
 		nil, nil, nil,
 		nopMutator,
 		nopMutator,
-		func() storage.Appender { return app },
+		func(ctx context.Context) storage.Appender { return app },
 		nil,
 		0,
 		true,
 	)
 
 	now := time.Now()
-	slApp := sl.appender()
+	slApp := sl.appender(context.Background())
 	_, _, _, err := sl.append(slApp, []byte("metric_a 1\n"), "", now)
 	testutil.Ok(t, err)
 	testutil.Ok(t, slApp.Commit())
 
-	slApp = sl.appender()
+	slApp = sl.appender(context.Background())
 	_, _, _, err = sl.append(slApp, []byte(""), "", now.Add(time.Second))
 	testutil.Ok(t, err)
 	testutil.Ok(t, slApp.Commit())
@@ -1232,19 +1471,19 @@ func TestScrapeLoopAppendNoStalenessIfTimestamp(t *testing.T) {
 		nil, nil, nil,
 		nopMutator,
 		nopMutator,
-		func() storage.Appender { return app },
+		func(ctx context.Context) storage.Appender { return app },
 		nil,
 		0,
 		true,
 	)
 
 	now := time.Now()
-	slApp := sl.appender()
+	slApp := sl.appender(context.Background())
 	_, _, _, err := sl.append(slApp, []byte("metric_a 1 1000\n"), "", now)
 	testutil.Ok(t, err)
 	testutil.Ok(t, slApp.Commit())
 
-	slApp = sl.appender()
+	slApp = sl.appender(context.Background())
 	_, _, _, err = sl.append(slApp, []byte(""), "", now.Add(time.Second))
 	testutil.Ok(t, err)
 	testutil.Ok(t, slApp.Commit())
@@ -1263,7 +1502,7 @@ func TestScrapeLoopRunReportsTargetDownOnScrapeError(t *testing.T) {
 	var (
 		scraper  = &testScraper{}
 		appender = &collectResultAppender{}
-		app      = func() storage.Appender { return appender }
+		app      = func(ctx context.Context) storage.Appender { return appender }
 	)
 
 	ctx, cancel := context.WithCancel(context.Background())
@@ -1291,7 +1530,7 @@ func TestScrapeLoopRunReportsTargetDownOnInvalidUTF8(t *testing.T) {
 	var (
 		scraper  = &testScraper{}
 		appender = &collectResultAppender{}
-		app      = func() storage.Appender { return appender }
+		app      = func(ctx context.Context) storage.Appender { return appender }
 	)
 
 	ctx, cancel := context.WithCancel(context.Background())
@@ -1345,14 +1584,14 @@ func TestScrapeLoopAppendGracefullyIfAmendOrOutOfOrderOrOutOfBounds(t *testing.T
 		nil, nil,
 		nopMutator,
 		nopMutator,
-		func() storage.Appender { return app },
+		func(ctx context.Context) storage.Appender { return app },
 		nil,
 		0,
 		true,
 	)
 
 	now := time.Unix(1, 0)
-	slApp := sl.appender()
+	slApp := sl.appender(context.Background())
 	total, added, seriesAdded, err := sl.append(slApp, []byte("out_of_order 1\namend 1\nnormal 1\nout_of_bounds 1\n"), "", now)
 	testutil.Ok(t, err)
 	testutil.Ok(t, slApp.Commit())
@@ -1377,7 +1616,7 @@ func TestScrapeLoopOutOfBoundsTimeError(t *testing.T) {
 		nil, nil,
 		nopMutator,
 		nopMutator,
-		func() storage.Appender {
+		func(ctx context.Context) storage.Appender {
 			return &timeLimitAppender{
 				Appender: app,
 				maxTime:  timestamp.FromTime(time.Now().Add(10 * time.Minute)),
@@ -1389,7 +1628,7 @@ func TestScrapeLoopOutOfBoundsTimeError(t *testing.T) {
 	)
 
 	now := time.Now().Add(20 * time.Minute)
-	slApp := sl.appender()
+	slApp := sl.appender(context.Background())
 	total, added, seriesAdded, err := sl.append(slApp, []byte("normal 1\n"), "", now)
 	testutil.Ok(t, err)
 	testutil.Ok(t, slApp.Commit())
@@ -1562,7 +1801,7 @@ func TestScrapeLoop_RespectTimestamps(t *testing.T) {
 	s := teststorage.New(t)
 	defer s.Close()
 
-	app := s.Appender()
+	app := s.Appender(context.Background())
 
 	capp := &collectResultAppender{next: app}
 
@@ -1570,13 +1809,13 @@ func TestScrapeLoop_RespectTimestamps(t *testing.T) {
 		nil, nil, nil,
 		nopMutator,
 		nopMutator,
-		func() storage.Appender { return capp },
+		func(ctx context.Context) storage.Appender { return capp },
 		nil, 0,
 		true,
 	)
 
 	now := time.Now()
-	slApp := sl.appender()
+	slApp := sl.appender(context.Background())
 	_, _, _, err := sl.append(slApp, []byte(`metric_a{a="1",b="1"} 1 0`), "", now)
 	testutil.Ok(t, err)
 	testutil.Ok(t, slApp.Commit())
@@ -1595,7 +1834,7 @@ func TestScrapeLoop_DiscardTimestamps(t *testing.T) {
 	s := teststorage.New(t)
 	defer s.Close()
 
-	app := s.Appender()
+	app := s.Appender(context.Background())
 
 	capp := &collectResultAppender{next: app}
 
@@ -1603,13 +1842,13 @@ func TestScrapeLoop_DiscardTimestamps(t *testing.T) {
 		nil, nil, nil,
 		nopMutator,
 		nopMutator,
-		func() storage.Appender { return capp },
+		func(ctx context.Context) storage.Appender { return capp },
 		nil, 0,
 		false,
 	)
 
 	now := time.Now()
-	slApp := sl.appender()
+	slApp := sl.appender(context.Background())
 	_, _, _, err := sl.append(slApp, []byte(`metric_a{a="1",b="1"} 1 0`), "", now)
 	testutil.Ok(t, err)
 	testutil.Ok(t, slApp.Commit())
@@ -1628,15 +1867,13 @@ func TestScrapeLoopDiscardDuplicateLabels(t *testing.T) {
 	s := teststorage.New(t)
 	defer s.Close()
 
-	app := s.Appender()
-
 	ctx, cancel := context.WithCancel(context.Background())
 	sl := newScrapeLoop(ctx,
 		&testScraper{},
 		nil, nil,
 		nopMutator,
 		nopMutator,
-		func() storage.Appender { return app },
+		s.Appender,
 		nil,
 		0,
 		true,
@@ -1644,7 +1881,7 @@ func TestScrapeLoopDiscardDuplicateLabels(t *testing.T) {
 	defer cancel()
 
 	// We add a good and a bad metric to check that both are discarded.
-	slApp := sl.appender()
+	slApp := sl.appender(ctx)
 	_, _, _, err := sl.append(slApp, []byte("test_metric{le=\"500\"} 1\ntest_metric{le=\"600\",le=\"700\"} 1\n"), "", time.Time{})
 	testutil.NotOk(t, err)
 	testutil.Ok(t, slApp.Rollback())
@@ -1656,7 +1893,7 @@ func TestScrapeLoopDiscardDuplicateLabels(t *testing.T) {
 	testutil.Ok(t, series.Err())
 
 	// We add a good metric to check that it is recorded.
-	slApp = sl.appender()
+	slApp = sl.appender(ctx)
 	_, _, _, err = sl.append(slApp, []byte("test_metric{le=\"500\"} 1\n"), "", time.Time{})
 	testutil.Ok(t, err)
 	testutil.Ok(t, slApp.Commit())
@@ -1673,10 +1910,10 @@ func TestScrapeLoopDiscardUnnamedMetrics(t *testing.T) {
 	s := teststorage.New(t)
 	defer s.Close()
 
-	app := s.Appender()
+	app := s.Appender(context.Background())
 
 	ctx, cancel := context.WithCancel(context.Background())
-	sl := newScrapeLoop(ctx,
+	sl := newScrapeLoop(context.Background(),
 		&testScraper{},
 		nil, nil,
 		func(l labels.Labels) labels.Labels {
@@ -1686,14 +1923,14 @@ func TestScrapeLoopDiscardUnnamedMetrics(t *testing.T) {
 			return l
 		},
 		nopMutator,
-		func() storage.Appender { return app },
+		func(ctx context.Context) storage.Appender { return app },
 		nil,
 		0,
 		true,
 	)
 	defer cancel()
 
-	slApp := sl.appender()
+	slApp := sl.appender(context.Background())
 	_, _, _, err := sl.append(slApp, []byte("nok 1\nnok2{drop=\"drop\"} 1\n"), "", time.Time{})
 	testutil.NotOk(t, err)
 	testutil.Ok(t, slApp.Rollback())
@@ -1793,6 +2030,7 @@ func TestReuseScrapeCache(t *testing.T) {
 		}
 		proxyURL, _ = url.Parse("http://localhost:2128")
 	)
+	defer sp.stop()
 	sp.sync([]*Target{t1})
 
 	steps := []struct {
@@ -1896,22 +2134,20 @@ func TestScrapeAddFast(t *testing.T) {
 	s := teststorage.New(t)
 	defer s.Close()
 
-	app := s.Appender()
-
 	ctx, cancel := context.WithCancel(context.Background())
 	sl := newScrapeLoop(ctx,
 		&testScraper{},
 		nil, nil,
 		nopMutator,
 		nopMutator,
-		func() storage.Appender { return app },
+		s.Appender,
 		nil,
 		0,
 		true,
 	)
 	defer cancel()
 
-	slApp := sl.appender()
+	slApp := sl.appender(ctx)
 	_, _, _, err := sl.append(slApp, []byte("up 1\n"), "", time.Time{})
 	testutil.Ok(t, err)
 	testutil.Ok(t, slApp.Commit())
@@ -1922,7 +2158,7 @@ func TestScrapeAddFast(t *testing.T) {
 		v.ref++
 	}
 
-	slApp = sl.appender()
+	slApp = sl.appender(ctx)
 	_, _, _, err = sl.append(slApp, []byte("up 1\n"), "", time.Time{}.Add(time.Second))
 	testutil.Ok(t, err)
 	testutil.Ok(t, slApp.Commit())
@@ -1947,6 +2183,7 @@ func TestReuseCacheRace(t *testing.T) {
 			},
 		}
 	)
+	defer sp.stop()
 	sp.sync([]*Target{t1})
 
 	start := time.Now()
