@@ -14,6 +14,7 @@
 package remote
 
 import (
+	"context"
 	"fmt"
 	"sync"
 	"time"
@@ -34,14 +35,6 @@ var (
 		Name:      "samples_in_total",
 		Help:      "Samples in to remote storage, compare to samples out for queue managers.",
 	})
-	highestTimestamp = maxGauge{
-		Gauge: promauto.NewGauge(prometheus.GaugeOpts{
-			Namespace: namespace,
-			Subsystem: subsystem,
-			Name:      "highest_timestamp_in_seconds",
-			Help:      "Highest timestamp that has come into the remote storage via the Appender interface, in seconds since epoch.",
-		}),
-	}
 )
 
 // WriteStorage represents all the remote write storage.
@@ -57,6 +50,10 @@ type WriteStorage struct {
 	queues            map[string]*QueueManager
 	samplesIn         *ewmaRate
 	flushDeadline     time.Duration
+	interner          *pool
+
+	// For timestampTracker.
+	highestTimestamp *maxGauge
 }
 
 // NewWriteStorage creates and runs a WriteStorage.
@@ -73,6 +70,18 @@ func NewWriteStorage(logger log.Logger, reg prometheus.Registerer, walDir string
 		flushDeadline:     flushDeadline,
 		samplesIn:         newEWMARate(ewmaWeight, shardUpdateDuration),
 		walDir:            walDir,
+		interner:          newPool(),
+		highestTimestamp: &maxGauge{
+			Gauge: prometheus.NewGauge(prometheus.GaugeOpts{
+				Namespace: namespace,
+				Subsystem: subsystem,
+				Name:      "highest_timestamp_in_seconds",
+				Help:      "Highest timestamp that has come into the remote storage via the Appender interface, in seconds since epoch.",
+			}),
+		},
+	}
+	if reg != nil {
+		reg.MustRegister(rws.highestTimestamp)
 	}
 	go rws.run()
 	return rws
@@ -149,6 +158,8 @@ func (rws *WriteStorage) ApplyConfig(conf *config.Config) error {
 			rwConf.WriteRelabelConfigs,
 			c,
 			rws.flushDeadline,
+			rws.interner,
+			rws.highestTimestamp,
 		)
 		// Keep track of which queues are new so we know which to start.
 		newHashes = append(newHashes, hash)
@@ -170,9 +181,10 @@ func (rws *WriteStorage) ApplyConfig(conf *config.Config) error {
 }
 
 // Appender implements storage.Storage.
-func (rws *WriteStorage) Appender() storage.Appender {
+func (rws *WriteStorage) Appender(_ context.Context) storage.Appender {
 	return &timestampTracker{
-		writeStorage: rws,
+		writeStorage:         rws,
+		highestRecvTimestamp: rws.highestTimestamp,
 	}
 }
 
@@ -187,9 +199,10 @@ func (rws *WriteStorage) Close() error {
 }
 
 type timestampTracker struct {
-	writeStorage     *WriteStorage
-	samples          int64
-	highestTimestamp int64
+	writeStorage         *WriteStorage
+	samples              int64
+	highestTimestamp     int64
+	highestRecvTimestamp *maxGauge
 }
 
 // Add implements storage.Appender.
@@ -212,7 +225,7 @@ func (t *timestampTracker) Commit() error {
 	t.writeStorage.samplesIn.incr(t.samples)
 
 	samplesIn.Add(float64(t.samples))
-	highestTimestamp.Set(float64(t.highestTimestamp / 1000))
+	t.highestRecvTimestamp.Set(float64(t.highestTimestamp / 1000))
 	return nil
 }
 
